# -*- coding: utf-8 -*-
import logging
import tempfile
from argparse import Namespace
from operator import itemgetter
from pathlib import Path
from tempfile import _TemporaryFileWrapper
from typing import List, Optional
from uuid import UUID

from apistar.exceptions import ErrorResponse
from arkindex_export import Element, open_database
from arkindex_export.queries import list_children
from arkindex_worker.cache import (
    CachedClassification,
    CachedElement,
    CachedEntity,
    CachedImage,
    CachedTranscription,
    CachedTranscriptionEntity,
    create_tables,
    create_version_table,
)
from arkindex_worker.cache import db as cache_database
from arkindex_worker.cache import init_cache_db
from arkindex_worker.image import download_image
from arkindex_worker.models import Dataset
from arkindex_worker.models import Element as WorkerElement
from arkindex_worker.utils import create_tar_zst_archive
from arkindex_worker.worker import DatasetWorker
from worker_generic_training_dataset.db import (
    list_classifications,
    list_transcription_entities,
    list_transcriptions,
    retrieve_element,
)
from worker_generic_training_dataset.utils import build_image_url

logger: logging.Logger = logging.getLogger(__name__)

BULK_BATCH_SIZE = 50
DEFAULT_TRANSCRIPTION_ORIENTATION = "horizontal-lr"


def _format_element(element: WorkerElement) -> Element:
    return retrieve_element(element.id)


class DatasetExtractor(DatasetWorker):
    def configure(self) -> None:
        self.args: Namespace = self.parser.parse_args()
        if self.is_read_only:
            super().configure_for_developers()
        else:
            super().configure()

        if self.user_configuration:
            logger.info("Overriding with user_configuration")
            self.config.update(self.user_configuration)

        # Download corpus
        self.download_latest_export()

    def configure_storage(self) -> None:
        self.data_folder = tempfile.TemporaryDirectory(suffix="-arkindex-data")
        self.data_folder_path = Path(self.data_folder.name)

        # Initialize db that will be written
        self.configure_cache()

        # CachedImage downloaded and created in DB
        self.cached_images = dict()

        # Where to save the downloaded images
        self.images_folder = self.data_folder_path / "images"
        self.images_folder.mkdir(parents=True)
        logger.info(f"Images will be saved at `{self.images_folder}`.")

    def configure_cache(self) -> None:
        """
        Create an SQLite database compatible with base-worker cache and initialize it.
        """
        self.use_cache = True
        self.cache_path: Path = self.data_folder_path / "db.sqlite"
        logger.info(f"Cached database will be saved at `{self.cache_path}`.")

        init_cache_db(self.cache_path)

        create_version_table()

        create_tables()

    def download_latest_export(self) -> None:
        """
        Download the latest export of the current corpus.
        Export must be in `"done"` state.
        """
        try:
            exports = list(
                self.api_client.paginate(
                    "ListExports",
                    id=self.corpus_id,
                )
            )
        except ErrorResponse as e:
            logger.error(
                f"Could not list exports of corpus ({self.corpus_id}): {str(e.content)}"
            )
            raise e

        # Find the latest that is in "done" state
        exports: List[dict] = sorted(
            list(filter(lambda exp: exp["state"] == "done", exports)),
            key=itemgetter("updated"),
            reverse=True,
        )
        assert (
            len(exports) > 0
        ), f"No available exports found for the corpus {self.corpus_id}."

        # Download latest export
        try:
            export_id: str = exports[0]["id"]
            logger.info(f"Downloading export ({export_id})...")
            self.export: _TemporaryFileWrapper = self.api_client.request(
                "DownloadExport",
                id=export_id,
            )
            logger.info(f"Downloaded export ({export_id}) @ `{self.export.name}`")
            open_database(self.export.name)
        except ErrorResponse as e:
            logger.error(
                f"Could not download export ({export_id}) of corpus ({self.corpus_id}): {str(e.content)}"
            )
            raise e

    def insert_classifications(self, element: CachedElement) -> None:
        logger.info("Listing classifications")
        classifications: list[CachedClassification] = [
            CachedClassification(
                id=classification.id,
                element=element,
                class_name=classification.class_name,
                confidence=classification.confidence,
                state=classification.state,
                worker_run_id=classification.worker_run,
            )
            for classification in list_classifications(element.id)
        ]
        if classifications:
            logger.info(f"Inserting {len(classifications)} classification(s)")
            with cache_database.atomic():
                CachedClassification.bulk_create(
                    model_list=classifications,
                    batch_size=BULK_BATCH_SIZE,
                )

    def insert_transcriptions(
        self, element: CachedElement
    ) -> List[CachedTranscription]:
        logger.info("Listing transcriptions")
        transcriptions: list[CachedTranscription] = [
            CachedTranscription(
                id=transcription.id,
                element=element,
                text=transcription.text,
                confidence=transcription.confidence,
                orientation=DEFAULT_TRANSCRIPTION_ORIENTATION,
                worker_version_id=transcription.worker_version,
                worker_run_id=transcription.worker_run,
            )
            for transcription in list_transcriptions(element.id)
        ]
        if transcriptions:
            logger.info(f"Inserting {len(transcriptions)} transcription(s)")
            with cache_database.atomic():
                CachedTranscription.bulk_create(
                    model_list=transcriptions,
                    batch_size=BULK_BATCH_SIZE,
                )
        return transcriptions

    def insert_entities(self, transcriptions: List[CachedTranscription]) -> None:
        logger.info("Listing entities")
        entities: List[CachedEntity] = []
        transcription_entities: List[CachedTranscriptionEntity] = []
        for transcription in transcriptions:
            for transcription_entity in list_transcription_entities(transcription.id):
                entity = CachedEntity(
                    id=transcription_entity.entity.id,
                    type=transcription_entity.entity.type.name,
                    name=transcription_entity.entity.name,
                    validated=transcription_entity.entity.validated,
                    metas=transcription_entity.entity.metas,
                    worker_run_id=transcription_entity.entity.worker_run,
                )
                entities.append(entity)
                transcription_entities.append(
                    CachedTranscriptionEntity(
                        id=transcription_entity.id,
                        transcription=transcription,
                        entity=entity,
                        offset=transcription_entity.offset,
                        length=transcription_entity.length,
                        confidence=transcription_entity.confidence,
                        worker_run_id=transcription_entity.worker_run,
                    )
                )
        if entities:
            # First insert entities since they are foreign keys on transcription entities
            logger.info(f"Inserting {len(entities)} entities")
            with cache_database.atomic():
                CachedEntity.bulk_create(
                    model_list=entities,
                    batch_size=BULK_BATCH_SIZE,
                )

        if transcription_entities:
            # Insert transcription entities
            logger.info(
                f"Inserting {len(transcription_entities)} transcription entities"
            )
            with cache_database.atomic():
                CachedTranscriptionEntity.bulk_create(
                    model_list=transcription_entities,
                    batch_size=BULK_BATCH_SIZE,
                )

    def insert_element(
        self, element: Element, parent_id: Optional[UUID] = None
    ) -> None:
        """
        Insert the given element in the cache database.
        Its image will also be saved to disk, if it wasn't already.

        The insertion of an element includes:
        - its classifications
        - its transcriptions
        - its transcriptions' entities (both Entity and TranscriptionEntity)

        :param element: Element to insert.
        :param parent_id: ID of the parent to use when creating the CachedElement. Do not specify for top-level elements.
        """
        logger.info(f"Processing element ({element.id})")
        if element.image and element.image.id not in self.cached_images:
            # Download image
            logger.info("Downloading image")
            download_image(url=build_image_url(element)).save(
                self.images_folder / f"{element.image.id}.jpg"
            )
            # Insert image
            logger.info("Inserting image")
            # Store images in case some other elements use it as well
            with cache_database.atomic():
                self.cached_images[element.image.id] = CachedImage.create(
                    id=element.image.id,
                    width=element.image.width,
                    height=element.image.height,
                    url=element.image.url,
                )

        # Insert element
        logger.info("Inserting element")
        with cache_database.atomic():
            cached_element: CachedElement = CachedElement.create(
                id=element.id,
                parent_id=parent_id,
                type=element.type,
                image=self.cached_images[element.image.id] if element.image else None,
                polygon=element.polygon,
                rotation_angle=element.rotation_angle,
                mirrored=element.mirrored,
                worker_version_id=element.worker_version,
                worker_run_id=element.worker_run,
                confidence=element.confidence,
            )

        # Insert classifications
        self.insert_classifications(cached_element)

        # Insert transcriptions
        transcriptions: List[CachedTranscription] = self.insert_transcriptions(
            cached_element
        )

        # Insert entities
        self.insert_entities(transcriptions)

    def process_split(self, split_name: str, elements: List[Element]) -> None:
        logger.info(
            f"Filling the cache with information from elements in the split {split_name}"
        )

        # First list all pages
        nb_elements: int = len(elements)
        for idx, element in enumerate(elements, start=1):
            logger.info(f"Processing `{split_name}` element ({idx}/{nb_elements})")

            # Insert page
            self.insert_element(element)

            # List children
            children = list_children(element.id)
            nb_children: int = children.count()
            for child_idx, child in enumerate(children, start=1):
                logger.info(f"Processing child ({child_idx}/{nb_children})")
                # Insert child
                self.insert_element(child, parent_id=element.id)

    def process_dataset(self, dataset: Dataset):
        # Configure temporary storage for the dataset data (cache + images)
        self.configure_storage()

        # Iterate over given splits
        for split_name, elements in self.list_dataset_elements_per_split(dataset):
            casted_elements = list(map(_format_element, elements))
            self.process_split(split_name, casted_elements)

        # TAR + ZSTD the cache and the images folder, and store as task artifact
        zstd_archive_path: Path = self.work_dir / f"{dataset.id}.zstd"
        logger.info(f"Compressing the images to {zstd_archive_path}")
        create_tar_zst_archive(
            source=self.data_folder_path, destination=zstd_archive_path
        )
        self.data_folder.cleanup()


def main():
    DatasetExtractor(
        description="Fill base-worker cache with information about dataset and extract images",
        generator=True,
    ).run()


if __name__ == "__main__":
    main()
